var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import PlaybackScheduler from "./PlaybackScheduler";
import { midiInstruments } from "./midi/midiInstruments";
import { SoundfontPlayer } from "./players/SoundfontPlayer";
import { getNoteDuration, getNoteVolume, getNoteArticulationStyle } from "./noteHelpers";
var PlaybackState;
(function (PlaybackState) {
    PlaybackState["INIT"] = "INIT";
    PlaybackState["PLAYING"] = "PLAYING";
    PlaybackState["STOPPED"] = "STOPPED";
    PlaybackState["PAUSED"] = "PAUSED";
})(PlaybackState || (PlaybackState = {}));
export default class PlaybackEngine {
    constructor() {
        this.defaultBpm = 100;
        this.availableInstruments = midiInstruments;
        this.scoreInstruments = [];
        this.ready = false;
        this.ac = new AudioContext();
        this.ac.suspend();
        this.instrumentPlayer = new SoundfontPlayer();
        this.instrumentPlayer.init(this.ac);
        this.cursor = null;
        this.sheet = null;
        this.denominator = null;
        this.scheduler = null;
        this.iterationSteps = 0;
        this.currentIterationStep = 0;
        this.timeoutHandles = [];
        this.playbackSettings = {
            bpm: this.defaultBpm,
            masterVolume: 1,
        };
        this.state = PlaybackState.INIT;
    }
    get wholeNoteLength() {
        return Math.round((60 / this.playbackSettings.bpm) * this.denominator * 1000);
    }
    getPlaybackInstrument(voiceId) {
        if (!this.sheet)
            return [0, ""];
        const voice = this.sheet.Instruments.flatMap(i => i.Voices).find(v => v.VoiceId === voiceId);
        return this.availableInstruments.find(i => i[0] === voice.midiInstrumentId);
    }
    setInstrument(voice, midiInstrumentId) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.instrumentPlayer.load(midiInstrumentId);
            voice.midiInstrumentId = midiInstrumentId;
        });
    }
    loadScore(osmd) {
        return __awaiter(this, void 0, void 0, function* () {
            this.ready = false;
            this.sheet = osmd.Sheet;
            this.scoreInstruments = this.sheet.Instruments;
            this.cursor = osmd.cursor;
            this.denominator = this.sheet.SheetPlaybackSetting.rhythm.Denominator;
            if (this.sheet.HasBPMInfo) {
                this.setBpm(this.sheet.DefaultStartTempoInBpm);
            }
            yield this.loadInstruments();
            this.initInstruments();
            this.scheduler = new PlaybackScheduler(this.denominator, this.wholeNoteLength, this.ac, (delay, notes) => this.notePlaybackCallback(delay, notes));
            this.countAndSetIterationSteps();
            this.ready = true;
            this.state = PlaybackState.STOPPED;
        });
    }
    initInstruments() {
        for (const i of this.sheet.Instruments) {
            for (const v of i.Voices) {
                v.midiInstrumentId = i.MidiInstrumentId;
            }
        }
    }
    loadInstruments() {
        return __awaiter(this, void 0, void 0, function* () {
            let playerPromises = [];
            for (const i of this.sheet.Instruments) {
                playerPromises.push(this.instrumentPlayer.load(i.MidiInstrumentId));
            }
            yield Promise.all(playerPromises);
        });
    }
    play() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.ac.resume();
            if (this.state === PlaybackState.INIT || this.state === PlaybackState.STOPPED) {
                this.cursor.show();
            }
            this.state = PlaybackState.PLAYING;
            this.scheduler.start();
        });
    }
    stop() {
        return __awaiter(this, void 0, void 0, function* () {
            this.state = PlaybackState.STOPPED;
            this.stopPlayers();
            this.clearTimeouts();
            this.scheduler.reset();
            this.cursor.reset();
            this.currentIterationStep = 0;
            this.cursor.hide();
        });
    }
    pause() {
        this.state = PlaybackState.PAUSED;
        this.ac.suspend();
        this.stopPlayers();
        this.scheduler.setIterationStep(this.currentIterationStep);
        this.scheduler.pause();
        this.clearTimeouts();
    }
    jumpToStep(step) {
        this.pause();
        if (this.currentIterationStep > step) {
            this.cursor.reset();
            this.currentIterationStep = 0;
        }
        while (this.currentIterationStep < step) {
            this.cursor.next();
            ++this.currentIterationStep;
        }
        let schedulerStep = this.currentIterationStep;
        if (this.currentIterationStep > 0 && this.currentIterationStep < this.iterationSteps)
            ++schedulerStep;
        this.scheduler.setIterationStep(schedulerStep);
    }
    setBpm(bpm) {
        this.playbackSettings.bpm = bpm;
        if (this.scheduler)
            this.scheduler.wholeNoteLength = this.wholeNoteLength;
    }
    countAndSetIterationSteps() {
        this.cursor.reset();
        let steps = 0;
        while (!this.cursor.Iterator.EndReached) {
            if (this.cursor.Iterator.CurrentVoiceEntries) {
                this.scheduler.loadNotes(this.cursor.Iterator.CurrentVoiceEntries);
            }
            this.cursor.next();
            ++steps;
        }
        this.iterationSteps = steps;
        this.cursor.reset();
    }
    notePlaybackCallback(audioDelay, notes) {
        if (this.state !== PlaybackState.PLAYING)
            return;
        let scheduledNotes = new Map();
        for (let note of notes) {
            const noteDuration = getNoteDuration(note, this.wholeNoteLength);
            if (noteDuration === 0)
                continue;
            const noteVolume = getNoteVolume(note);
            const noteArticulation = getNoteArticulationStyle(note);
            const midiPlaybackInstrument = note.ParentVoiceEntry.ParentVoice.midiInstrumentId;
            const fixedKey = note.ParentVoiceEntry.ParentVoice.Parent.SubInstruments[0].fixedKey || 0;
            if (!scheduledNotes.has(midiPlaybackInstrument)) {
                scheduledNotes.set(midiPlaybackInstrument, []);
            }
            scheduledNotes.get(midiPlaybackInstrument).push({
                note: note.halfTone - fixedKey * 12,
                duration: noteDuration / 1000,
                gain: noteVolume,
                articulation: noteArticulation,
            });
        }
        for (const [midiId, notes] of scheduledNotes) {
            this.instrumentPlayer.schedule(midiId, this.ac.currentTime + audioDelay, notes);
        }
        this.timeoutHandles.push(setTimeout(() => this.iterationCallback(), Math.max(0, audioDelay * 1000 - 40))); // Subtracting 40 milliseconds to compensate for update delay
    }
    stopPlayers() {
        for (const i of this.sheet.Instruments) {
            for (const v of i.Voices) {
                this.instrumentPlayer.stop(v.midiInstrumentId);
            }
        }
    }
    // Used to avoid duplicate cursor movements after a rapid pause/resume action
    clearTimeouts() {
        for (let h of this.timeoutHandles) {
            clearTimeout(h);
        }
        this.timeoutHandles = [];
    }
    iterationCallback() {
        if (this.state !== PlaybackState.PLAYING)
            return;
        if (this.currentIterationStep > 0)
            this.cursor.next();
        ++this.currentIterationStep;
    }
}
